from skimage import io, color
from PIL import Image

import numpy as np
import cv2

from scipy.optimize import minimize
from scipy.optimize import Bounds

from utils import *

import warnings
warnings.filterwarnings("ignore", category=RuntimeWarning)

def correction_( L0_, shadow, highlight, shadow_tone, highlight_tone ):
    '''
    ### formula: f(s, h, st, ht) ###
    
    Given:
        `L0_`: A original luminance at pixel location (i, j); (range: [0, 1])
        `shadow_map`: A value of shadow weights at pixel location (i, j) in shadow map
        `highlight_map`: A value of highlight weights at pixel location (i, j) in highlight map
        `shadow`: A shadow number (default: 0)  (range: [-inf, inf])
        `highlight`: A highlight number (default: 0) (range: [-inf, inf])

    Return:
        `L_correct`: A w-by-h array of corrected luminance
    '''
    
    tone_to_map = lambda t, lum: 1 / (1 + np.exp( 25 * ( t - lum ) ) )
    shadow_map = tone_to_map( shadow_tone, L0_ )
    highlight_map = tone_to_map( highlight_tone, 1 - L0_ )
    
    # apply correction on shadow and highlight (use some scaling to avoid fast trasition)
    correct = lambda lum, opt: - lum / ( np.exp( - opt / 10 ) * ( lum - 1 ) - lum )
    
    # feathering & reconstruction
    L_correct = ( 1 - shadow_map ) * L0_ + shadow_map * correct( L0_, shadow )
    L_correct = ( 1 - highlight_map ) * L_correct + highlight_map * correct( L_correct, highlight )
    return L_correct

def color_optimization( palette, weights, target_color, verbose=False ):
    '''
    Given:
        `palette`: A #p-by-2 array
        `weights`: A array of shape (#p,)
        `target_colors`: A array of shape (2,)
        `verbose`: toggle for displaying optimization results
    
    Return:
        `palette_opt`: A #p-by-2 array from optimization results
    '''
    
    assert palette.shape[0] == weights.shape[0]
    assert palette.shape[1] == target_color.shape[0]
    
    ### objective function
    def func( vec_delp ):
        delp = vec_delp.reshape( palette.shape )
        rec_colors = weights @ ( palette + delp )
        
        sparsity = sum( np.linalg.norm( delp, axis = 1 ) )     # optimize for local palette's sparsity
        color_cons = np.linalg.norm( rec_colors - target_color )
        return sparsity + 3 * color_cons
    
    bounds = Bounds( -128 - palette.flatten(), 127 - palette.flatten() )
    x0 = np.zeros( palette.shape[0] * palette.shape[1] )
    res = minimize( func, x0, method='SLSQP', options={'ftol': 1e-6, 'disp': False, 'maxiter': 300}, bounds=bounds )
    delp = ( res.x ).reshape( palette.shape )
    
    if verbose:
        print( '-- Original weights:', weights )
        print( '-- Original color: ', weights @ palette )
        print( '-- Target color: ', target_color )
        print( '-- Reconstructed color: ', weights @ ( delp + palette ) )
    
    return delp + palette

def lum_optimization( L0, L_target, i, j, sh_palette, weights, verbose=False ):
    '''
    Given:
        `L0`: A w-by-h array of original luminance (range: [0, 1])
        `L_target`: A float value of target luminance (range: [0, 1])
        `i`,`j`: Pixel location 
        `sh_palette`: A #P-by-2 shadow/highlight palette
        `weights`: A array of shape (#p,)
        `verbose`: toggle for displaying optimization results
    
    Return:
        `sh_opt`: An #p-by-4 shadow/highlight control array
    '''
    
    assert sh_palette.shape[0] == weights.shape[0]
    
    L0_at_pixel = L0[i, j]
    
    def compute_luminance( sh_opt ):
        lum_sum = 0
        for i in range( sh_opt.shape[0] ):
            lum_sum += weights[i] * correction_( L0_at_pixel, sh_opt[i, 0], sh_opt[i, 1], sh_opt[i, 2], sh_opt[i, 3] )
        return lum_sum
    
    def func( sh ):
        sh_ = sh.reshape( sh_palette.shape )
        err = ( compute_luminance( sh_ ) - L_target ) ** 2          # error term
        per = sum( np.linalg.norm( sh_ - sh_palette, axis = 1 ) )
        return err + 0.0001 * per 
    
    ## set boundary
    row = [ (-200, 200), (-200, 200), (0., 1.), (0., 1.) ]
    bnds = [ bnd for i in range( sh_palette.shape[0] ) for bnd in row ]
    
    ## convert boundaries to constraints
    ## from: https://stackoverflow.com/questions/12781622/does-scipys-minimize-function-with-method-cobyla-accept-bounds
    cons = []
    for factor in range( len( bnds ) ):
        lower, upper = bnds[factor]
        l = { 'type': 'ineq', 'fun': lambda x, lb=lower, i=factor: x[i] - lb }
        u = { 'type': 'ineq', 'fun': lambda x, ub=upper, i=factor: ub - x[i] }
        cons.append( l )
        cons.append( u )
    
    ### optimization
    sh_init = np.ones( sh_palette.shape[0] * sh_palette.shape[1] )
    res = minimize( func, sh_init, method='SLSQP', constraints=cons, options={'ftol': 1e-8, 'disp': False, 'maxiter': 1000} )
    sh_opt = ( res.x ).reshape( sh_palette.shape )
    
    if verbose:
        np.set_printoptions(suppress=True)
        recon_lum = compute_luminance( sh_opt )
        print( res )
        print( '-- Original lum: ', L0_at_pixel )
        print( '-- Target lum: ', L_target )
        print( '-- Error: ', np.abs( recon_lum - L_target ) )
        print( '-- Temp reconstructed lum (w/o computing tone): ', recon_lum ) 
        print( '-- Optimized s/h palette (Ms, Mh):\n ', sh_opt )
    
        #sh_opt[:, 2], sh_opt[:, 3] =  tone_convert_func( L0_at_pixel, sh_opt[:, 2], sh_opt[:, 3] )
    return sh_opt
    
def dynamic_function_plot( sh_palette, weights, L0 ):
    '''
    Given:
        `sh_palette`: A #P-by-2 shadow/highlight palette
        `weights`: A array of shape (#p,)
        `L0`: A w-by-h array of original luminance (range: [0, 1])
    '''
        
    def func( sh_palette, weights, L ):
        lum_range = np.zeros_like( L )
        for i in range( sh_palette.shape[0] ):
            lum_range += weights[i] * correction_( L, sh_palette[i, 0], sh_palette[i, 1], sh_palette[i, 2], sh_palette[i, 3] )
        return lum_range
    
    import matplotlib.pyplot as plt
    
    L = np.linspace( 0, 1, 500 )
    L_range = func( sh_palette, weights, L )
    
    fig, ax = plt.subplots()
    ax.plot( L, L_range, linewidth = 2.0 )
    ax.set( xlim = ( 0, 1 ), ylim = ( 0, 1 ) )
    plt.show()
    #return L, L_range


def main():
    import argparse
    parser = argparse.ArgumentParser( description = 'shadow/highlight testing.' )
    parser.add_argument( 'img', help = 'The path to the input image.' )
    args = parser.parse_args()
    
    img = color.rgb2lab( np.array( Image.open( args.img ) ) / 255. )   
    ab_palette = extract_AB_palette( img )
    weights = extract_weights( ab_palette, img ).reshape( (img.shape[0], img.shape[1], 5) )
    
    L0 = img[:, :, 0] / 100
    ab = weights @ ab_palette
    
    target_color = np.array( [-35, -10] )
    target_lum = 0.7
    
    i, j = 100, 52
    sh_palette = np.zeros( (5,4) )
    p_weights = weights[i, j]
    
    print( 'Weights: ', p_weights )
    
    ### color optimization
    #print( 'Original color: ', ab[i, j] )
    #print( 'Target color: ', target_color )
    #palette_opt = color_optimization( ab_palette, p_weights, target_color )
    
    ### luminance optimization
    print( 'Original lum: ', L0[i,j] )
    print( 'Target lum: ', target_lum )
    sh_opt = lum_optimization( L0, target_lum, i, j, sh_palette, p_weights, verbose=True )
    print( sh_opt )
    #weights_ = np.zeros_like( p_weights )
    #weights_[-1] = 1
    dynamic_function_plot( sh_opt, p_weights, L0 )
    
    ### reconstruction
    recon_lum = get_recon_lum( sh_opt, weights.reshape( -1, 5 ), L0 * 100 )
    recon_img = get_recon_img( ab_palette, weights, recon_lum )
    print( 'Final reconstructed lum: ', recon_lum[i,j] / 100 )
    
    ### save result
    recon_img = ( color.lab2rgb( recon_img ) * 255. ).clip( 0, 255 ).astype( np.uint8 )
    Image.fromarray( recon_img ).save( '../test-edit.png' )
    

if __name__ == '__main__':
    main()