import numpy as np
import tensorflow as tf

from scipy.optimize import minimize
from scipy.optimize import Bounds

from skimage import io, color
from PIL import Image

try:
    from utils import *
except ImportError:
    from .utils import *
    
import warnings
warnings.filterwarnings("ignore", category=RuntimeWarning)

N = 100
w_c = 100
num_p = 5

def grad( n ):
    G = np.zeros( ( n-1, n ) )
    for i in range( 0, n-1 ):
        G[i,i+1] = 1
        G[i,i] = -1
    return G

def objective_func( L, constraints, W ):
    # building bilaplacian
    G = grad( N )
    Lap = G.T @ G
    
    # mirroring the laplacian to avoid prefering 0 gradients at endpoints
    Lap[0,:] = 0
    Lap[-1,:] = 0
    B = Lap.T @ Lap
    
    # building selection and constraint matrices
    C = np.zeros( ( N, len( constraints ) ) )
    S = np.zeros( ( N, len( constraints ) ) )
    for i, ( Lindex, val ) in enumerate( constraints ):
        C[Lindex, i] = val
        S[Lindex, i] = 1
    
    LBBL = L.T @ B @ L
    
    trace = 0
    for i in range( LBBL.shape[0] ):
        trace += LBBL[i, i]
    return trace + w_c * np.sum( (np.multiply( S, L @ W ) - C)**2 )

def linear_sys_solve( constraints, W ):
    assert W.shape[0] == num_p
    assert W.shape[1] == len( constraints )
    
    # identities in kronecker product
    id_p = np.eye( num_p ) 
    id_n = np.eye( N ) 
    
    # building bilaplacian
    G = grad( N )
    Lap = G.T @ G
    
    # mirroring the laplacian to avoid prefering 0 gradients at endpoints
    Lap[0,:] = 0
    Lap[-1,:] = 0
    B = Lap
    
    # building selection and constraint matrices
    C = np.zeros( ( N, len( constraints ) ) )
    S = np.zeros( ( N, len( constraints ) ) )
    for i, ( Lindex, val ) in enumerate( constraints ):
        C[Lindex, i] = val
        S[Lindex, i] = 1
    
    
    # compute left-hand and right-hand side
    # note: `flatten('F')` is vectorization on columns
    W_idn_kron = np.kron( W, id_n )
    A = np.kron( id_p, 2*B.T@B ) + w_c * W_idn_kron @ np.diag( (2*np.multiply(S,S)).flatten('F') ) @ np.kron( W.T, id_n )
    b = w_c * ( W_idn_kron @ ( np.multiply( 2*C, S ).flatten('F') ) )
    
    # change the rows to identity for known values
    id_p_n = np.eye( num_p * N )
    zero_lum_ind = [N*i for i in range(num_p)]
    one_lum_ind = [N*(i+1)-1 for i in range(num_p)]
    
    A[zero_lum_ind, :] = id_p_n[zero_lum_ind, :]
    A[one_lum_ind, :] = id_p_n[one_lum_ind, :]
    b[zero_lum_ind] = 0
    b[one_lum_ind] = 1
    
    x_l = np.linalg.solve( A, b )
    L_ = x_l.reshape( N, num_p, order='F' )
    
    return L_

def compute_new_luminance( L0, L_bilap, weights ):
    from scipy import interpolate
    
    L_new = 0
    L = np.linspace( 0, 1, N )
    for i in range( weights.shape[1] ):
        f = interpolate.interp1d( L, L_bilap[:,i] )
        L_new += np.multiply( f( L0 ), weights[:, i] )
    return L_new

def function_plot_for_curves( L_bilap, constraints, L0, W ):
    import matplotlib.pyplot as plt
    
    L = np.linspace( 0, 1, N )
    
    fig, ax = plt.subplots( nrows = 2, ncols = 3 )
    fig.tight_layout( pad = 1.0 )
    axes = ax.flatten()
    
    L_ = 0
    for i in range( num_p+1 ):
        if i == num_p:
            axes[i].set_title( 'Entire function', fontdict = {'fontsize': 7} )
            axes[i].plot( L, L_bilap@W , label = "Bilaplacian" )
        else:
            #f = interpolate.interp1d( L, L_bilap[:,i] )
            #L_ += np.multiply( f( L0 ), weights[:, i] )
            
            axes[i].set_title( 'Palette ' + str(i), fontdict = {'fontsize': 7} )
            axes[i].plot( L, L_bilap[:,i], label = "Bilaplacian" )
            
        axes[i].plot( L, L, label = "Benchmark" )
        axes[i].axvline( x = constraints[0][0]/N, linewidth = 1.5, color = 'r', label = 'L0', linestyle=':' )
        axes[i].axhline( y = constraints[0][1], linewidth = 1.5, color = 'g', label = 'L0\'', linestyle=':' )
        axes[i].legend( loc="lower right", prop={'size': 5} )
    plt.show()


def main():
    import argparse
    parser = argparse.ArgumentParser( description = 'shadow/highlight testing.' )
    parser.add_argument( 'img', help = 'The path to the input image.' )
    args = parser.parse_args()
    np.set_printoptions( suppress = True )
    
    img = color.rgb2lab( np.array( Image.open( args.img ) ) / 255. )   
    ab_palette = extract_AB_palette( img, 4 )   # palette size here
    weights = extract_weights( ab_palette, img ).reshape( (img.shape[0], img.shape[1], ab_palette.shape[0]) )
    
    # get constraint data
    L0 = img[:, :, 0] / 100
    ab = weights @ ab_palette
    
    # pixel coords
    i, j = 100, 52
    i2, j2 = 30, 250
    
    target_lum = 0.3
    target_lum2 = 0.5
    
    #W = np.array( [weights[i, j]] ).T
    W = np.array( [weights[i, j], weights[i2, j2]] ).T
    #constraints = [ ( int(L0[i, j]*N), target_lum ) ]
    constraints = [ ( int(L0[i, j]*N), target_lum ), ( int(L0[i2, j2]*N), target_lum2 ) ]
    
    # scale down the gray and re-normalize weights
    W[-1, :] /= 100
    W = W / np.sum(W, axis=0)
    
    #print( 'Luminance constraints:\n', (L0[i, j], target_lum) )
    print( 'Luminance constraints:\n', (L0[i, j], target_lum), (L0[i2, j2], target_lum2) )
    print( 'Weights at constraints:\n', W )
    
    L_ = linear_sys_solve( constraints, W )
    
    show_verify = 1
    if show_verify:
        import mygrad as mg
        L = mg.tensor(L_)
        loss = objective_func( L, constraints, W )
        
        #L_bench = np.tile( np.array([np.linspace( 0, 1, N )]).transpose(), (1, num_p) )
        #print( objective_func( L_bench, constraints, W ) )
        
        #print( objective_func( np.linspace( 0, 1, N ), constraints, W ) )
        loss.backward()
        print( 'Norm of gradient at the solution: ', np.linalg.norm( L.grad.data, 'fro' ) )
    
    ### reconstruction
    weights = weights.reshape( -1, num_p )
    weights[:, -1] /= 100
    weights = (weights.T / np.sum(weights, axis=1)).T
    
    recon_lum = (compute_new_luminance( L0.reshape(-1), L_, weights ) * 100).reshape( img.shape[0], img.shape[1] )
    
    #print( 'Reconstructed luminance:\n', recon_lum[i, j]/100 )
    print( 'Reconstructed luminance:\n', recon_lum[i, j]/100, recon_lum[i2, j2]/100 )
    
    img[:, :, 0] = recon_lum
    img = ( color.lab2rgb( img ) * 255. ).clip( 0, 255 ).astype( np.uint8 )
    Image.fromarray( img ).save( '../test-edit-bilap.png' )
    
    function_plot_for_curves( L_, constraints, L0.reshape(-1), W )
    
if __name__ == '__main__':
    main()